"""
This module defines an UFL transformation, that takes a UFL expression
and transforms it into a sum of terms that all contain the correct number
of test function terms (which is equal to the form rank).
"""
from __future__ import absolute_import

from dune.perftool.ufl.modified_terminals import extract_modified_arguments
from dune.perftool.ufl.transformations import ufl_transformation
from dune.perftool.ufl.transformations.replace import replace_expression

from ufl.algorithms import MultiFunction
from ufl.classes import Zero, IntValue

import itertools


class _ReplacementDict(dict):
    def __init__(self, *mod_args):
        dict.__init__(self)
        for ma in mod_args:
            self[ma] = IntValue(1)

    def __getitem__(self, key):
        return self.get(key, Zero())


@ufl_transformation(name="accterms2", extraction_lambda=lambda l: [i[0] for i in l])
def split_into_accumulation_terms(expr):
    mod_args = extract_modified_arguments(expr)

    accumulation_terms = []

    # Treat the case of a rank 1 form:
    if len(filter(lambda ma: ma.argexpr.count() == 1, mod_args)) == 0:
        for arg in mod_args:
            # Do the replacement on the expression
            accum_expr = replace_expression(expr, replacemap=_ReplacementDict(arg.expr))

            # Store the found accumulation expression
            accumulation_terms.append((accum_expr, (arg,)))
    # and now the case of a rank 2 form:
    else:
        for arg1, arg2 in itertools.product(filter(lambda ma: ma.argexpr.count() == 0, mod_args),
                                            filter(lambda ma: ma.argexpr.count() == 1, mod_args)
                                            ):
            accum_expr = replace_expression(expr, replacemap=_ReplacementDict(arg1.expr, arg2.expr))

            accumulation_terms.append((accum_expr, (arg1, arg2)))

    # and return the result
    return accumulation_terms
